from types import FunctionType

from django.conf import settings
from django.contrib import messages
from django.templatetags.l10n import localize
from django.utils.html import strip_spaces_between_tags
from django.utils.translation import (
    get_language_from_request, get_language_info)
from jinja2.environment import Environment
from jinja2.ext import Extension
from jinja2.nodes import Assign, CallBlock, ContextReference, Name, Output
from rjsmin import jsmin


from ..strings import dash_case_convert


class LanguageInfoExtension(Extension):
    """
    An assigment tag for Jinja, setting a language info dictionary.

    Samples hacked together from:
    http://stackoverflow.com/a/23170408/1067833
    https://github.com/coffin/coffin/blob/master/coffin/static.py
    """

    tags = set(['get_my_language_info'])

    def parse(self, parser):
        stream = parser.stream
        lineno = next(stream).lineno

        ctx_ref = ContextReference()
        call_node = self.call_method(
            '_get_current_language_info', [ctx_ref], lineno=lineno)

        if stream.next_if('name:as'):
            var = Name(stream.expect('name').value, 'store')
            return Assign(var, call_node).set_lineno(lineno)
        else:
            return Output([call_node]).set_lineno(lineno)

    def _get_current_language_info(self, context):
        lang_code = \
            get_language_from_request(request=context['request']) \
            or settings.LANGUAGE_CODE
        lang_info = get_language_info(lang_code=lang_code)
        lang_info['locale_code'] = lang_code
        return lang_info


class SpacelessExtension(Extension):
    """
    Removes whitespace between HTML tags, including tab and
    newline characters.

    Works exactly like Django's own tag.
    """

    tags = set(['spaceless'])

    def parse(self, parser):
        lineno = next(parser.stream).lineno
        body = parser.parse_statements(
            end_tokens=('name:endspaceless',), drop_needle=True)
        return CallBlock(
            self.call_method('_strip_spaces', [], [], None, None),
            [], [], body,
        ).set_lineno(lineno)

    def _strip_spaces(self, caller=None):
        return strip_spaces_between_tags(caller().strip())


class JsMinExtension(Extension):
    """
    Removes whitespace between JavaScript tags, including tab and
    newline characters.
    """

    tags = {'jsmin'}

    def parse(self, parser):
        lineno = parser.stream.__next__().lineno
        body = parser.parse_statements(
            end_tokens=('name:endjsmin',), drop_needle=True)
        return CallBlock(
            self.call_method('_strip_whitespace_js', [], [], None, None),
            [], [], body,
        ).set_lineno(lineno)

    def _strip_whitespace_js(self, caller: FunctionType):
        if settings.DEBUG:
            # Debug mode, no minification
            return caller().strip()
        return jsmin(caller().strip())

# https://github.com/coffin/coffin/blob/master/coffin/common.py


class KToolsDjangoExtension(Extension):
    'Extensions/utils that are also used from within this module.'

    def __init__(self, environment: Environment):
        super().__init__(environment=environment)
        environment.globals['dash_case_convert'] = dash_case_convert
        environment.globals['messages'] = messages
        environment.filters['localize'] = localize
