from django.conf import settings
from django.core.handlers.wsgi import WSGIRequest
from django.http.response import JsonResponse
from django.middleware.csrf import rotate_token
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_protect
from django.views.generic.base import ContextMixin, TemplateView, View


class JsonResponseMixin(ContextMixin, View):
    """
    A mixin that can be used to render a JSON response. Override
    `get_json_data()` and you're good to go.
    """
    status_code = 200

    def render_to_response(
            self, context: dict, status_code: int | None = None,
            **response_kwargs) -> JsonResponse:
        """
        Returns a JSON response, transforming 'context' to make the
        payload.
        """
        json_dumps_params = {}
        if settings.DEBUG:
            json_dumps_params['indent'] = 2
        return JsonResponse(
            data=self.get_json_data(context=context),
            status=status_code or self.status_code,
            json_dumps_params=json_dumps_params,
            **response_kwargs
        )

    def get_json_data(self, context: dict) -> dict:
        """
        Returns an object that will be serialized as JSON by
        `json.dumps()`.
        """
        return {}

    get = TemplateView.get


@method_decorator(decorator=csrf_protect, name='dispatch')
class RotateTokenMixin(View):
    'Enforce and rotate CSRF token on each request.'

    def dispatch(self, request: WSGIRequest, *args, **kwargs):
        'Rotate the token here.'
        rotate_token(request=request)
        return super().dispatch(request, *args, **kwargs)
