__all__ = ['TempLock']


from fcntl import LOCK_EX, LOCK_UN, flock
from os import getuid
from tempfile import gettempdir, gettempprefix

from .pathlib import Path

_KTOOLS_TEMPLOCK = Path(
    gettempdir(), f'{gettempprefix()}-ktools-lockfile-{getuid()}')
_KTOOLS_TEMPLOCK.touch(exist_ok=True)


class _KLock(object):
    """
    A context-using file locker. HEADS UP: man 2 flock:

    Locks are on files, not file descriptors.  That is, file descriptors
    duplicated through dup(2) or fork(2) do not result in multiple
    instances of a lock, but rather multiple references to a single
    lock. If a process holding a lock on a file forks and the child
    explicitly unlocks the file, the parent will lose its lock.

    Processes blocked awaiting a lock may be awakened by signals.
    """
    path = _KTOOLS_TEMPLOCK

    def __enter__(self):
        'Enter the lock.'
        self.file_descriptor = open(self.path, 'r')
        flock(self.file_descriptor, LOCK_EX)
        return self.file_descriptor

    def __exit__(self, exc_type, exc_value, traceback):
        'Exit the lock.'
        flock(self.file_descriptor, LOCK_UN)
        self.file_descriptor.close()


class TempLock(object):
    """
    A lock that will create the lock file for itself, in a thread-safe
    manner.
    """
    _is_acquired = False

    def __init__(self, path: Path):
        self._path = path

    def __enter__(self):
        'Enter the lock.'
        with _KLock():
            self._path.touch(exist_ok=True)
        self._file = self._path.open(mode='r')
        flock(self._file, LOCK_EX)
        self._is_acquired = True

    def __exit__(self, exc_type, exc_value, traceback):
        'Exit the lock.'
        if not self._is_acquired:
            return
        flock(self._file, LOCK_UN)
        self._file.close()
        self._is_acquired = False

    acquire = __enter__

    def release(self):
        'Release the lock.'
        return self.__exit__(exc_type=None, exc_value=None, traceback=None)
